/* $Id$ */

#include "image/lkTracker.hpp"

using namespace jafar::image;

LKTracker::LKTracker() :
  qualityLevel(0.01), minDistance(10), harrisFreeParam(0.04), halfWinSize(10), maxPyramidLevel(3),
  m_nbPoints(0), m_maxNbPoints(500), m_pyramidFlags(0),
  m_imagePrev(0),
  m_pointsPrev(0), m_pointsCur(0),
  m_status(0), m_errors(0)
{}

LKTracker::~LKTracker()
{
  if (m_imagePrev)   delete m_imagePrev;
  if (!m_pointsPrev.empty())  m_pointsPrev.clear();
  if (!m_pointsCur.empty())   m_pointsCur.clear();
  if (!m_status.empty())      m_status.clear();
  if (!m_errors.empty())      m_errors.clear();
}

void LKTracker::init(Image const& image)
{
  JFR_PRECOND(image.colorSpace() == JfrImage_CS_GRAY,
	      "LKTracker::init: invalid image");

  if (m_imagePrev)   delete m_imagePrev;
  if (!m_pointsPrev.empty())  m_pointsPrev.clear();
  if (!m_pointsCur.empty())   m_pointsCur.clear();
  if (!m_status.empty())      m_status.clear();
  if (!m_errors.empty())      m_errors.clear();
  m_pyramidFlags = 0;

  // memory allocation
  m_imagePrev   = new Image(image.width(), image.height(), CV_8U, JfrImage_CS_GRAY);
  m_pointsPrev  = std::vector<cv::Point2f>(m_maxNbPoints);
  m_pointsCur   = std::vector<cv::Point2f>(m_maxNbPoints);
  m_status      = std::vector<unsigned char>(m_maxNbPoints);
  m_errors      = std::vector<float>(m_maxNbPoints);
  // points detection
  //m_nbPoints = m_maxNbPoints;

  cv::goodFeaturesToTrack(image, m_pointsCur, m_maxNbPoints, qualityLevel, minDistance,
			  cv::Mat(), 3, true, harrisFreeParam);

  cv::cornerSubPix(image, m_pointsCur, cv::Size(halfWinSize, halfWinSize), cv::Size(-1,-1),
		   cv::TermCriteria(cv::TermCriteria::MAX_ITER|cv::TermCriteria::EPS,20,0.03));
  //  JFR_DEBUG("nbPoints: " << m_nbPoints);
  image.copy(*m_imagePrev);
}

void LKTracker::track(Image const& image)
{
  JFR_PRECOND(image.colorSpace() == JfrImage_CS_GRAY,
	      "LKTracker::init: invalid image");
  
  if (m_nbPoints > 0) {
    // swap
    m_pointsPrev.swap(m_pointsCur);
  
    // track points
    // cvCalcOpticalFlowPyrLK(*m_imagePrev, image, *m_pyramidPrev, *m_pyramidCur,
    // 			   m_pointsPrev, m_pointsCur, m_nbPoints, cvSize(halfWinSize, halfWinSize), 
    // 			   maxPyramidLevel, m_status, 0,
    // 			   cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.03), m_pyramidFlags);
    cv::calcOpticalFlowPyrLK(*m_imagePrev, image, 
			     m_pointsPrev, m_pointsCur, 
			     m_status, m_errors,
			     cv::Size(halfWinSize, halfWinSize), maxPyramidLevel,
			     cv::TermCriteria(cv::TermCriteria::MAX_ITER|cv::TermCriteria::EPS,20,0.03), 
			     0, m_pyramidFlags);

    m_pyramidFlags |= CV_LKFLOW_PYR_A_READY;

    // remove lost points
    int i, k;
    for( i = k = 0; i < m_nbPoints; ++i ) {
      if( !m_status[i] )
	continue;
      m_pointsCur[k++] = m_pointsCur[i];
    }
    m_nbPoints = k;
    //    JFR_DEBUG("nbPoints: " << m_nbPoints);
    image.copy(*m_imagePrev);
  }
}
