
#include "image/Image.hpp"

#include <stdint.h>

namespace jafar {
  namespace image {
    namespace details {
      inline double pow2(double a) { return a * a; }
      template<int depth, typename worktype, worktype borneinf, worktype bornesup>
      double computeZncc(Image const* im1, Image const* im2, cv::Rect const* roiRect = 0)
      {
	JFR_PRECOND( im1->depth() == depth, "Image 1 depth is different from the template parameter" );
	JFR_PRECOND( im2->depth() == depth, "Image 2 depth is different from the template parameter" );
	JFR_PRECOND( im1->channels() == im2->channels(), "The channels number of both images are different" );
	JFR_PRECOND( im1->height() == im2->height(), "The height of both images are different" );
	JFR_PRECOND( im1->width() == im2->width(), "The width of both images are different" );
			
	int roiX, roiY, roiWidth, roiHeight, roiStep;
	if (roiRect)
	  {
	    roiX = roiRect->x;
	    roiY = roiRect->y;
	    roiWidth = roiRect->width;
	    roiHeight = roiRect->height;
	    if (roiX+roiWidth > im1->width())
	      {
		roiWidth = im1->width()-roiX;
		//JFR_WARNING("ROI too large, internally changed the width");
	      }
	    if (roiY+roiHeight > im1->height())
	      {
		roiHeight = im1->height() - roiY;
		//JFR_WARNING("ROI too large, internally changed the height");
	      }
	    roiStep = im1->step() - roiWidth;
	  }
	else
	  {
	    roiX = 0;
	    roiY = 0;
	    roiWidth = im1->width();
	    roiHeight = im1->height();
	    roiStep = im1->step() - roiWidth;
	  }
			
	// bring the iterator to the right position
	worktype const* im1ptr = reinterpret_cast<worktype const*>(im1->data());
	worktype const* im2ptr = reinterpret_cast<worktype const*>(im2->data());
	im1ptr += roiY*im1->step()+roiX;
	im2ptr += roiY*im2->step()+roiX;
			
	//       int height = im1->height();
	//       int width = im1->width();
	//       int step = im1->step() - width;
	int zncc_count = 0;
    
	double mean1 = 0.;
	double mean2 = 0.;
	// Compute the mean
	for(int i=roiY;i<(roiY+roiHeight);i++) {
	  for(int j=roiX;j<(roiX+roiWidth);j++) {
	    worktype im1v = *(im1ptr++);
	    worktype im2v = *(im2ptr++);
	    if(im1v != borneinf && im1v != bornesup && im2v != borneinf && im2v != bornesup )
	      {
		mean1 += im1v;
		mean2 += im2v;
		zncc_count++;
	      }
	  }
	  im1ptr += roiStep;
	  im2ptr += roiStep;
	}
	mean1 /= zncc_count;
	mean2 /= zncc_count; 
    
	// Compute the sigma and the zncc sum
			
	// bring the iterator to the right position
	im1ptr = reinterpret_cast<worktype const*>(im1->data());
	im2ptr = reinterpret_cast<worktype const*>(im2->data());
	im1ptr += roiY*im1->step()+roiX;
	im2ptr += roiY*im2->step()+roiX;
			
	double sigma1 = 0.;
	double sigma2 = 0.;
	double zncc_sum = 0.;
	for(int i=roiY;i < (roiY+roiHeight);i++) {
	  for(int j=roiX;j < (roiX+roiWidth);j++) {
	    worktype im1v = *(im1ptr++);
	    worktype im2v = *(im2ptr++);
	    if(im1v != borneinf && im1v != bornesup && im2v != borneinf && im2v != bornesup )
	      {
		sigma1 += details::pow2( im1v - mean1 );
		sigma2 += details::pow2( im2v - mean2 );
		zncc_sum += im1v * im2v;
	      }
	  }
	  im1ptr += roiStep;
	  im2ptr += roiStep;
	}
	//       JFR_DEBUG(sigma1 << " " << sigma2 << " " << mean1 << " " << mean2 << " " << zncc_count);
	sigma1 = sqrt(sigma1 / zncc_count);
	sigma2 = sqrt(sigma2 / zncc_count);
	zncc_sum = zncc_sum / zncc_count - mean1 * mean2;
    
	//std::cout << " zncc:: " << sigma1 << " " << sigma2 << " " << zncc_sum << std::endl;
			
	return zncc_sum / (sigma1 * sigma2);
      } 

      template<int depth, typename worktype, worktype borneinf, worktype bornesup>
      double computeZncc3(Image const* im1, Image const* im2, cv::Rect const* roi1_ = 0, cv::Rect const* roi2_ = 0)
      {
	JFR_PRECOND( im1->depth() == depth, "Image 1 depth is different from the template parameter" );
	JFR_PRECOND( im2->depth() == depth, "Image 2 depth is different from the template parameter" );
	JFR_PRECOND( im1->channels() == im2->channels(), "The channels number of both images are different" );
			
	cv::Rect roi1, roi2;
	if (roi1_) roi1 = *roi1_; else { roi1.x=0; roi1.y=0; roi1.width=im1->width(); roi1.height=im1->height(); }
	if (roi1_) roi2 = *roi2_; else { roi2.x=0; roi2.y=0; roi2.width=im1->width(); roi2.height=im1->height(); }
			
	JFR_PRECOND( roi1.width  == roi2.width , "The width of both images or roi are different" );
	JFR_PRECOND( roi1.height == roi2.height, "The height of both images or roi are different" );

	unsigned roi1_step, roi2_step;
	roi1_step = im1->step() - roi1.width;
	roi2_step = im2->step() - roi2.width;
			
	// Compute the mean
	worktype const* im1ptr = reinterpret_cast<worktype const*>(im1->data());
	worktype const* im2ptr = reinterpret_cast<worktype const*>(im2->data());
	im1ptr += roi1.y*im1->step()+roi1.x;
	im2ptr += roi2.y*im2->step()+roi2.x;
			
	double mean1 = 0.;
	double mean2 = 0.;
	int zncc_count = 0;
			
	for(int i = 0; i < roi1.height; ++i) 
	  {
	    for(int j = 0; j < roi1.width; ++j) 
	      {
		worktype im1v = *(im1ptr++);
		worktype im2v = *(im2ptr++);
		if(im1v != borneinf && im1v != bornesup && im2v != borneinf && im2v != bornesup )
		  {
		    mean1 += im1v;
		    mean2 += im2v;
		    zncc_count++;
		  }
	      }
	    im1ptr += roi1_step;
	    im2ptr += roi2_step;
	  }
	mean1 /= zncc_count;
	mean2 /= zncc_count; 
		
	// Compute the sigma and the zncc sum
	im1ptr = reinterpret_cast<worktype const*>(im1->data());
	im2ptr = reinterpret_cast<worktype const*>(im2->data());
	im1ptr += roi1.y*im1->step()+roi1.x;
	im2ptr += roi2.y*im2->step()+roi2.x;
			
	double sigma1 = 0.;
	double sigma2 = 0.;
	double zncc_sum = 0.;
			
	for(int i = 0; i < roi1.height; ++i) 
	  {
	    for(int j = 0; j < roi1.width; ++j) 
	      {
		worktype im1v = *(im1ptr++);
		worktype im2v = *(im2ptr++);
		if(im1v != borneinf && im1v != bornesup && im2v != borneinf && im2v != bornesup )
		  {
		    sigma1 += details::pow2( im1v - mean1 );
		    sigma2 += details::pow2( im2v - mean2 );
		    zncc_sum += im1v * im2v;
		  }
	      }
	    im1ptr += roi1_step;
	    im2ptr += roi2_step;
	  }
	sigma1 = sqrt(sigma1 / zncc_count);
	sigma2 = sqrt(sigma2 / zncc_count);
	zncc_sum = zncc_sum / zncc_count - mean1 * mean2;
			
	return zncc_sum / (sigma1 * sigma2);
      } 



      // a bit faster, but a bit less precise (but it should be ok even for a full image)
      template<int depth, typename worktype, worktype borneinf, worktype bornesup, bool useWeightMatrix>
      double computeZnccRoisWeight(Image const* im1, Image const* im2, cv::Rect const* roi1_ = NULL, cv::Rect const* roi2_ = NULL, float const* weightMatrix = NULL)
      {
	JFR_PRECOND( im1->depth() == depth, "Image 1 depth is different from the template parameter" );
	JFR_PRECOND( im2->depth() == depth, "Image 2 depth is different from the template parameter" );
	JFR_PRECOND( im1->channels() == im2->channels(), "The channels number of both images are different" );
	JFR_PRECOND( useWeightMatrix == (weightMatrix != NULL), "Template parameter tells to use weightMatrix but no one is given" );
			
	cv::Rect roi1, roi2;
	if (roi1_) roi1 = *roi1_; else { roi1.x=0; roi1.y=0; roi1.width=im1->width(); roi1.height=im1->height(); }
	if (roi1_) roi2 = *roi2_; else { roi2.x=0; roi2.y=0; roi2.width=im2->width(); roi2.height=im2->height(); }
			
	JFR_PRECOND( roi1.width  == roi2.width , "The width of both images or roi are different" );
	JFR_PRECOND( roi1.height == roi2.height, "The height of both images or roi are different" );

	unsigned roi1_step, roi2_step;
	roi1_step = im1->step() - roi1.width;
	roi2_step = im2->step() - roi2.width;
			
	double mean1 = 0., mean2 = 0.;
	double sigma1 = 0., sigma2 = 0.;
	double zncc_sum = 0.;
	double zncc_count = 0.;
	double zncc_total = 0.;
			
	worktype const* im1ptr = reinterpret_cast<worktype const*>(im1->data());
	worktype const* im2ptr = reinterpret_cast<worktype const*>(im2->data());
	im1ptr += roi1.y*im1->step()+roi1.x;
	im2ptr += roi2.y*im2->step()+roi2.x;
	float const* wptr = weightMatrix;
	double w;
			
	for(int i = 0; i < roi1.height; ++i) 
	  {
	    for(int j = 0; j < roi1.width; ++j) 
	      {
		worktype im1v = *(im1ptr++);
		worktype im2v = *(im2ptr++);
		if (useWeightMatrix) w = *(wptr++); else w = 1;
		zncc_total += w;
					
		if(im1v != borneinf && im1v != bornesup && im2v != borneinf && im2v != bornesup )
		  {
		    /*double im1vw, im2vw;
		    if (useWeightMatrix)
		      {
			im1vw = im1v * w;
			im2vw = im2v * w;
		      } else
		      {
			im1vw = im1v;
			im2vw = im2v;
		      }
		    */						
		    zncc_count += w;
		    mean1 += im1v * w;
		    mean2 += im2v * w;
		    sigma1 += im1v * im1v * w;
		    sigma2 += im2v * im2v * w;
		    zncc_sum += im1v * im2v * w;
		  }
	      }
	    im1ptr += roi1_step;
	    im2ptr += roi2_step;
	  }
			
	if (zncc_count / zncc_total < 0.5) return 0;
	//JFR_DEBUG("zncc_count " << zncc_count << ", zncc_total " << zncc_total << ", zncc_sum " << zncc_sum << ", mean1 " << mean1 << ", mean2 " << mean2 << ", sigma1 " << sigma1 << ", sigma2 " << sigma2);
			
	mean1 /= zncc_count;
	mean2 /= zncc_count;
	sigma1 = sqrt(sigma1/zncc_count - mean1*mean1);
	sigma2 = sqrt(sigma2/zncc_count - mean2*mean2);
	//JFR_DEBUG("zncc_count " << zncc_count << ", zncc_total " << zncc_total << ", zncc_sum " << zncc_sum << ", mean1 " << mean1 << ", mean2 " << mean2 << ", sigma1 " << sigma1 << ", sigma2 " << sigma2);
	zncc_sum = (zncc_sum/zncc_count - mean1*mean2) / (sigma1*sigma2);
			
	return zncc_sum;
      }

#if 1
      double computeZnccFloat(Image const* im1, Image const* im2)
      {
	JFR_PRECOND( im1->depth() == CV_32F, "Image 1 depth is different from the template parameter" );
	JFR_PRECOND( im2->depth() == CV_32F, "Image 2 depth is different from the template parameter" );
	JFR_PRECOND( im1->channels() == im2->channels(), "The channels number of both images are different" );
	JFR_PRECOND( im1->height() == im2->height(), "The height of both images are different" );
	JFR_PRECOND( im1->width() == im2->width(), "The width of both images are different" );
  
  
	int height = im1->height();
	int width = im1->width();
  
	int zncc_count = 0;
  
	double mean1 = 0.;
	double mean2 = 0.;
        // Compute the mean
	for(int i=0;i < height;i++) {
	  float const* im1ptr = reinterpret_cast<float const*>( im1->data(i) );
	  float const* im2ptr = reinterpret_cast<float const*>( im2->data(i) );
	  for(int j=0;j < width;j++) {
	    float im1v = *(im1ptr++);
	    float im2v = *(im2ptr++);
	    //       JFR_DEBUG(i << j << " " << im1v << " " << im2v);
	    mean1 += im1v;
	    mean2 += im2v;
	    zncc_count++;
	    //         JFR_DEBUG(i << j << " " << mean1 << " " << mean2 << " " << zncc_count);
	  }
	}
	mean1 /= zncc_count;
	mean2 /= zncc_count;
  
	// Compute the sigma and the zncc sum
	double sigma1 = 0.;
	double sigma2 = 0.;
	double zncc_sum = 0.;
	for(int i=0;i < height;i++) {
	  float const* im1ptr = reinterpret_cast<float const*>( im1->data(i) );
	  float const* im2ptr = reinterpret_cast<float const*>( im2->data(i) );
	  for(int j=0;j < width;j++) {
	    float im1v = *(im1ptr++);
	    float im2v = *(im2ptr++);
	    sigma1 += details::pow2( im1v - mean1 );
	    sigma2 += details::pow2( im2v - mean2 );
	    zncc_sum += im1v * im2v;
	    //       JFR_DEBUG(i << j << " " << sigma1 << " " << sigma2 << " " << zncc_sum);
	  }
	}
	//   JFR_DEBUG(sigma1 << " " << sigma2 << " " << mean1 << " " << mean2 << " " << zncc_count);
	sigma1 = sqrt(sigma1 / zncc_count);
	sigma2 = sqrt(sigma2 / zncc_count);
	zncc_sum = zncc_sum / zncc_count - mean1 * mean2;
  
	return zncc_sum / (sigma1 * sigma2);
      }
#endif

    }
    double Image::computeZncc(Image const* im2 ) const
    {
      JFR_PRECOND(this->depth() == im2->depth(), "The depth of both images is different");
      switch(this->depth())
	{
//	case CV_1U:
//	  return details::computeZncc<CV_1U, bool,0,1>(this,im2);
	case CV_8U:
	  return details::computeZncc<CV_8U, uint8_t,0,255>(this,im2);
	case CV_8S:
	  return details::computeZncc<CV_8S, int8_t, -127,128>(this,im2);
	case CV_16U:
	  return details::computeZncc<CV_16U, uint16_t, 0,65535>(this,im2);
	case CV_16S:
	  return details::computeZncc<CV_16S, int16_t, -32767,32768>(this,im2);
	case CV_32F:
	  return details::computeZnccFloat(this,im2);
	default:
	  JFR_PRECOND(false, "Unknown image depth");
	  return FP_NAN;
	}
    }

    double Image::computeZnccRoisWeight(Image const* im2, cv::Rect const* roi1_, cv::Rect const* roi2_, float const* weightMatrix) const
    {
      JFR_PRECOND(this->depth() == im2->depth(), "The depth of both images is different");
      switch(this->depth())
	{
//	case CV_1U:
//	  if (weightMatrix == NULL)
//	    return details::computeZnccRoisWeight<CV_1U, bool,0,1,false>(this,im2, roi1_, roi2_);
//	  else
//	    return details::computeZnccRoisWeight<CV_1U, bool,0,1,true>(this,im2, roi1_, roi2_,weightMatrix);
	case CV_8U:
	  if (weightMatrix == NULL)
	    return details::computeZnccRoisWeight<CV_8U, uint8_t,0,255,false>(this,im2, roi1_, roi2_);
	  else
	    return details::computeZnccRoisWeight<CV_8U, uint8_t,0,255,true>(this,im2, roi1_, roi2_,weightMatrix);
	case CV_8S:
	  if (weightMatrix == NULL)
	    return details::computeZnccRoisWeight<CV_8S, int8_t, -127,128,false>(this,im2, roi1_, roi2_);
	  else
	    return details::computeZnccRoisWeight<CV_8S, int8_t, -127,128,true>(this,im2, roi1_, roi2_,weightMatrix);
	case CV_16U:
	  if (weightMatrix == NULL)
	    return details::computeZnccRoisWeight<CV_16U, uint16_t, 0,65535,false>(this,im2, roi1_, roi2_);
	  else
	    return details::computeZnccRoisWeight<CV_16U, uint16_t, 0,65535,true>(this,im2, roi1_, roi2_,weightMatrix);
	case CV_16S:
	  if (weightMatrix == NULL)
	    return details::computeZnccRoisWeight<CV_16S, int16_t, -32767,32768,false>(this,im2, roi1_, roi2_);
	  else
	    return details::computeZnccRoisWeight<CV_16S, int16_t, -32767,32768,true>(this,im2, roi1_, roi2_,weightMatrix);
	  /*		case CV_32F:
			if (weightMatrix == NULL)
			return details::computeZnccRoisWeight<CV_32F, float, 0.,0.,false>(this,im2, roi1_, roi2_);
			else
			return details::computeZnccRoisWeight<CV_32F, float, 0.,0.,true>(this,im2, roi1_, roi2_,weightMatrix);
	  */		default:
	  JFR_PRECOND(false, "Unknown image depth");
	  return FP_NAN;
	}
    }


    double Image::computeZnccRoi(Image const* im2, cv::Rect const* rectRoi) const
    {
      JFR_PRECOND(this->depth() == im2->depth(), "The depth of both images is different");
      switch(this->depth())
	{
//	case CV_1U:
//	  return details::computeZncc<CV_1U, bool,0,1>(this,im2,rectRoi);
	case CV_8U:
	  return details::computeZncc<CV_8U, uint8_t,0,255>(this,im2,rectRoi);
	case CV_8S:
	  return details::computeZncc<CV_8S, int8_t, -127,128>(this,im2,rectRoi);
	case CV_16U:
	  return details::computeZncc<CV_16U, uint16_t, 0,65535>(this,im2,rectRoi);
	case CV_16S:
	  return details::computeZncc<CV_16S, int16_t, -32767,32768>(this,im2,rectRoi);
	case CV_32F:
	  JFR_PRECOND(false, "ZNCC calculation is not yet supported for images with CV_32F");
	default:
	  JFR_PRECOND(false, "Unknown image depth");
	}
      return -1;
    }

    double Image::computeRotatingZncc( Image const* im2, int rotationStep) const
    {
      // const IplImage* ipl2 = (const IplImage*) im2;
      double bestZncc = this->computeZncc(im2);
      //   double tempBestAngle = 0.;
      for(int angle = rotationStep; angle < 360; angle += rotationStep)
	{
	  float m[6];
	  int w = im2->width();
	  int h = im2->height();
	  float radangle = angle * CV_PI/180.;
	  m[0] = (float)(cos(radangle));
	  m[1] = (float)(sin(radangle));
	  m[2] = w * 0.5f;
	  m[3] = -m[1];
	  m[4] = m[0];
	  m[5] = h * 0.5f;
	  image::Image* im2bis = new image::Image( im2->width(), im2->height(), im2->depth(), im2->colorSpace());
	  // IplImage* ipl2bis = (IplImage*) im2bis;
	  // CvMat M = cvMat( 2, 3, CV_32F, m );
	  // cvGetQuadrangleSubPix( ipl2, ipl2bis, &M);
	  cv::Mat M = cv::Mat( 2, 3, CV_32FC1, m );
	  getQuadrangleSubPix( *im2, *im2bis, M);
	  double zncc = this->computeZncc(im2bis);
	  if(zncc > bestZncc) { bestZncc = zncc; /*tempBestAngle = radangle;*/}
	  delete im2bis;
	}
      //   JFR_DEBUG(tempBestAngle);
      return bestZncc;
    }

  }
}
